/*
  ktracker (c) 2006 Kris Beevers

  This file is part of ktracker.

  ktracker is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ktracker is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ktracker; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

  $Id: video_output.hpp,v 1.1.1.1 2006/10/10 20:41:29 beevek Exp $
*/

#ifndef _VIDEO_OUTPUT_HPP
#define _VIDEO_OUTPUT_HPP

// rudimetary ffmpeg based video output.  doesn't actually work very
// well because i haven't gotten around to figuring out how to set the
// PTS of a frame properly just yet.

#include <inttypes.h>
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>

// codec_id comes from ffmpeg/avcodec.h
// some useful codecs:
//   CODEC_ID_MPEG1VIDEO
//   CODEC_ID_MPEG4
//   CODEC_ID_HUFFYUV  (lossless)
//   CODEC_ID_FFV1     (lossless)

class video_output
{
public:
  video_output() : init_okay(false), did_alloc(false) {}
  ~video_output() { close(); }

  // you must call one of these before calling open
  void set_context
    (uint32_t w, uint32_t h, uint32_t bit_rate = 400000,
     uint32_t dt_num = 1, uint32_t dt_den = 25, uint32_t frames_between_intra = 10);
  void set_context(AVCodecContext *context) { ctx = context; }

  // you can either pass in a context from, e.g., video_input, or you
  // can make one with make_context, or you can make one yourself
  bool open(const char *output, CodecID codec_id = CODEC_ID_MPEG1VIDEO);
  void close();
  bool encode_frame(const uint8_t *bgr, uint32_t time_ms = 0);

protected:
  bool init_okay;
  AVCodec *codec;
  AVCodecContext *ctx;
  bool did_alloc;
  AVFrame *frame, *frame_bgr;
  FILE *file;
  uint32_t c_id;
  uint8_t *buf;
  uint32_t buf_sz, out_size;
};

void video_output::set_context
  (uint32_t w, uint32_t h, uint32_t bit_rate,
   uint32_t dt_num, uint32_t dt_den, uint32_t frames_between_intra)
{
  ctx = avcodec_alloc_context();
  did_alloc = true;
  ctx->bit_rate = bit_rate;
  ctx->width = w;
  ctx->height = h;
  ctx->time_base.num = dt_num;
  ctx->time_base.den = dt_den;
  ctx->gop_size = frames_between_intra;
  ctx->pix_fmt = PIX_FMT_YUV420P;
}

bool video_output::open(const char *output, CodecID codec_id)
{
  if(!ctx)
    return false;

  // open the codec
  codec = avcodec_find_encoder(codec_id);
  if(!codec)
    return false;
  if(avcodec_open(ctx, codec) < 0)
    return false;
  c_id = codec_id;

  // open the output file
  file = fopen(output, "wb");
  if(!file)
    return false;

  // allocate output buffer
  buf_sz = ctx->width*ctx->height*3 * 3; // 3 images of space
  buf = new uint8_t[buf_sz];
  out_size = 0;

  frame = avcodec_alloc_frame();
  uint32_t sz = avpicture_get_size(PIX_FMT_YUV420P, ctx->width, ctx->height);
  uint8_t *fbuf = new uint8_t[sz];
  avpicture_fill((AVPicture *)frame, fbuf, PIX_FMT_YUV420P, ctx->width, ctx->height);

  frame_bgr = avcodec_alloc_frame();

  init_okay = true;
  return true;
}

void video_output::close()
{
  if(!init_okay)
    return;

  // flush delayed frames
  while(out_size) {
    out_size = avcodec_encode_video(ctx, buf, buf_sz, 0);
    fwrite(buf, 1, out_size, file);
  }

  if(c_id == CODEC_ID_MPEG1VIDEO || c_id == CODEC_ID_MPEG2VIDEO ||
     c_id == CODEC_ID_MPEG4) { // it's an mpeg, write sequence end code
    buf[0] = 0x00;
    buf[1] = 0x00;
    buf[2] = 0x01;
    buf[3] = 0xb7;
    fwrite(buf, 1, 4, file);
  }

  delete [] buf;
  avcodec_close(ctx);
  av_free(frame);
  if(did_alloc)
    av_free(ctx);
  fclose(file);
  init_okay = false;
}

bool video_output::encode_frame(const uint8_t *bgr, uint32_t time_ms)
{
  if(!init_okay)
    return false;

  // FIXME FIXME: why doesn't this work to scale the framerate appropriately?
  if(time_ms != 0) {
    const AVRational bq = {1,1000}; // convert ms to 1/25ths of a sec
    frame->pts = av_rescale_q(time_ms, bq, ctx->time_base);
  }

  // use img_convert from libavcodec to convert from bgr to yuv420p,
  // then output the yuv420p image
  avpicture_fill((AVPicture *)frame_bgr, (uint8_t *)bgr,
		 PIX_FMT_BGR24, ctx->width, ctx->height);
  img_convert((AVPicture *)frame, PIX_FMT_YUV420P,
	      (AVPicture *)frame_bgr, PIX_FMT_BGR24,
	      ctx->width, ctx->height);

  out_size = avcodec_encode_video(ctx, buf, buf_sz, frame);
  fwrite(buf, 1, out_size, file);
  return true;
}

#endif // _VIDEO_OUTPUT_HPP
